@@ -41,6 +41,7 @@
 #include "einfo.h"
 #include "queue.h"
 #include "rc.h"
+#include "rc_exec.h"
 #include "rc-logger.h"
 #include "misc.h"
 #include "plugin.h"
@@ -64,14 +65,12 @@ const char * const longopts_help[] = {
 	"output the RC system type, if any",
 	longopts_help_COMMON
 };
-const char *usagestring = ""					\
-    "Usage: openrc [options] [<runlevel>]";
+const char *usagestring = ""
+	"Usage: openrc [options] [<runlevel>]";
 
 #define INITSH                  RC_LIBEXECDIR "/sh/init.sh"
 #define INITEARLYSH             RC_LIBEXECDIR "/sh/init-early.sh"
 
-#define INTERACTIVE             RC_SVCDIR "/interactive"
-
 #define DEVBOOT			"/dev/.rcboot"
 
 const char *applet = NULL;
@@ -94,29 +93,25 @@ clean_failed(void)
 {
 	DIR *dp;
 	struct dirent *d;
-	char *path;
 
 	/* Clean the failed services state dir now */
-	if ((dp = opendir(RC_SVCDIR "/failed"))) {
-		while ((d = readdir(dp))) {
-			if (d->d_name[0] == '.' &&
-			    (d->d_name[1] == '\0' ||
-				(d->d_name[1] == '.' && d->d_name[2] == '\0')))
-				continue;
+	if (!(dp = do_opendirat(rc_dirfd(RC_DIR_SVCDIR), "failed")))
+		return;
 
-			xasprintf(&path, RC_SVCDIR "/failed/%s", d->d_name);
-			if (unlink(path))
-				eerror("%s: unlink `%s': %s",
-				    applet, path, strerror(errno));
-			free(path);
-		}
-		closedir(dp);
+	while ((d = readdir(dp))) {
+		if (strcmp(d->d_name, ".") == 0 || strcmp(d->d_name, "..") == 0)
+			continue;
+
+		if (unlinkat(dirfd(dp), d->d_name, 0) == -1)
+			eerror("%s: unlink '%s/failed/%s': %s", applet, rc_svcdir(), d->d_name, strerror(errno));
 	}
+	closedir(dp);
 }
 
 static void
 cleanup(void)
 {
+	int svcdirfd = rc_dirfd(RC_DIR_SVCDIR);
 	RC_PID *p, *tmp;
 
 	if (!rc_in_logger && !rc_in_plugin &&
@@ -133,8 +128,9 @@ cleanup(void)
 		}
 
 		/* Clean runlevel start, stop markers */
-		rmdir(RC_STARTING);
-		rmdir(RC_STOPPING);
+		unlinkat(svcdirfd, "rc.starting", AT_REMOVEDIR);
+		unlinkat(svcdirfd, "rc.stopping", AT_REMOVEDIR);
+
 		clean_failed();
 		rc_logger_close();
 	}
@@ -211,8 +207,9 @@ want_interactive(void)
 static void
 mark_interactive(void)
 {
-	FILE *fp = fopen(INTERACTIVE, "w");
-	if (fp)
+	FILE *fp;
+
+	if ((fp = do_fopenat(rc_dirfd(RC_DIR_SVCDIR), "interactive", O_WRONLY | O_CREAT | O_TRUNC)))
 		fclose(fp);
 }
 
@@ -335,11 +332,11 @@ static char *get_krunlevel(void)
 		return NULL;
 	}
 
-	if (getline(&buffer, &i, fp) != -1) {
-		i = strlen(buffer);
-		if (buffer[i - 1] == '\n')
-			buffer[i - 1] = 0;
+	if (xgetline(&buffer, &i, fp) == -1) {
+		free(buffer);
+		buffer = NULL;
 	}
+
 	fclose(fp);
 	return buffer;
 }
@@ -486,17 +483,19 @@ do_sysinit(void)
 		run_program(INITEARLYSH);
 
 	uname(&uts);
-	printf("\n   %sOpenRC %s" VERSION "%s is starting up %s",
-	    ecolor(ECOLOR_GOOD), ecolor(ECOLOR_HILITE),
-	    ecolor(ECOLOR_NORMAL), ecolor(ECOLOR_BRACKET));
+	if (!rc_yesno(getenv("EINFO_QUIET"))) {
+		printf("\n   %sOpenRC %s" VERSION "%s is starting up %s",
+		    ecolor(ECOLOR_GOOD), ecolor(ECOLOR_HILITE),
+		    ecolor(ECOLOR_NORMAL), ecolor(ECOLOR_BRACKET));
 #ifdef BRANDING
-	printf(BRANDING " (%s)", uts.machine);
+		printf(BRANDING " (%s)", uts.machine);
 #else
-	printf("%s %s (%s)",
-	    uts.sysname,
-	    uts.release,
-	    uts.machine);
+		printf("%s %s (%s)",
+		    uts.sysname,
+		    uts.release,
+		    uts.machine);
 #endif
+	}
 
 	if ((sys = rc_sys()))
 		printf(" [%s]", sys);
@@ -515,8 +514,7 @@ do_sysinit(void)
 	 * sys */
 	if ((sys = rc_sys()))
 		setenv("RC_SYS", sys, 1);
-	/* force an update of the dependency tree */
-	if ((main_deptree = _rc_deptree_load(1, NULL)) == NULL)
+	if ((main_deptree = _rc_deptree_load(false, NULL)) == NULL)
 		eerrorx("failed to load deptree");
 }
 
@@ -527,7 +525,7 @@ runlevel_config(const char *service, const char *level)
 	char *conf, *dir;
 	bool retval;
 
-	dir = dirname(init);
+	dir = dirname(dirname(init));
 	xasprintf(&conf, "%s/conf.d/%s.%s", dir, service, level);
 	retval = exists(conf);
 	free(conf);
@@ -543,12 +541,13 @@ do_stop_services(RC_STRINGLIST *types_nw, RC_STRINGLIST *start_services,
 	pid_t pid;
 	RC_STRING *service, *svc1, *svc2;
 	RC_STRINGLIST *deporder, *tmplist, *kwords;
+	RC_STRINGLIST *types_nw_save = NULL;
 	RC_SERVICE state;
 	RC_STRINGLIST *nostop;
 	bool crashed, nstop;
 
 	if (!types_nw) {
-		types_nw = rc_stringlist_new();
+		types_nw = types_nw_save = rc_stringlist_new();
 		rc_stringlist_add(types_nw, "needsme");
 		rc_stringlist_add(types_nw, "wantsme");
 	}
@@ -638,6 +637,9 @@ stop:
 		}
 	}
 
+	if (types_nw_save)
+		rc_stringlist_free(types_nw_save);
+
 	rc_stringlist_free(nostop);
 }
 
@@ -651,7 +653,7 @@ do_start_services(const RC_STRINGLIST *start_services, bool parallel)
 	bool crashed = false;
 
 	if (!rc_yesno(getenv("EINFO_QUIET")))
-		interactive = exists(INTERACTIVE);
+		interactive = existsat(rc_dirfd(RC_DIR_SVCDIR), "interactive");
 	errno = 0;
 	crashed = rc_conf_yesno("rc_crashed_start");
 	if (errno == ENOENT)
@@ -706,14 +708,11 @@ do_start_services(const RC_STRINGLIST *start_services, bool parallel)
 
 	/* Store our interactive status for boot */
 	if (interactive &&
-	    (strcmp(runlevel, RC_LEVEL_SYSINIT) == 0 ||
-		strcmp(runlevel, getenv("RC_BOOTLEVEL")) == 0))
+			(strcmp(runlevel, RC_LEVEL_SYSINIT) == 0 ||
+			 strcmp(runlevel, getenv("RC_BOOTLEVEL")) == 0))
 		mark_interactive();
-	else {
-		if (exists(INTERACTIVE))
-			unlink(INTERACTIVE);
-	}
-
+	else
+		unlinkat(rc_dirfd(RC_DIR_SVCDIR), "interactive", 0);
 }
 
 #ifdef RC_DEBUG
@@ -753,6 +752,7 @@ int main(int argc, char **argv)
 	bool going_down = false;
 	int depoptions = RC_DEP_STRICT | RC_DEP_TRACE;
 	char *krunlevel = NULL;
+	const char *workingdir = "/";
 	char *pidstr = NULL;
 	int opt;
 	bool parallel;
@@ -773,26 +773,15 @@ int main(int argc, char **argv)
 	applet = basename_c(argv[0]);
 	LIST_INIT(&service_pids);
 	LIST_INIT(&free_these_pids);
-	atexit(cleanup);
 	if (!applet)
 		eerrorx("arguments required");
 
 	argc--;
 	argv++;
 
-	/* Change dir to / to ensure all scripts don't use stuff in pwd */
-	if (chdir("/") == -1)
-		eerror("chdir: %s", strerror(errno));
-
-	/* Ensure our environment is pure
-	 * Also, add our configuration to it */
-	env_filter();
-	env_config();
-
 	/* complain about old configuration settings if they exist */
 	if (exists(RC_CONF_OLD)) {
-		ewarn("%s still exists on your system and should be removed.",
-				RC_CONF_OLD);
+		ewarn("%s still exists on your system and should be removed.", RC_CONF_OLD);
 		ewarn("Please migrate to the appropriate settings in %s", RC_CONF);
 	}
 
@@ -837,19 +826,35 @@ int main(int argc, char **argv)
 		}
 	}
 
-	if (strcmp(applet, "rc") == 0)
-		ewarn("rc is deprecated, please use openrc instead.");
+	/* Change dir to / to ensure all init scripts don't use stuff in pwd
+	 * For user services, change to the user's HOME instead. */
+	if (rc_is_user() && !(workingdir = getenv("HOME")))
+		eerrorx("HOME is unset.");
+
+	if (chdir(workingdir) == -1)
+		eerror("chdir: %s", strerror(errno));
+
+	/* Ensure our environment is pure
+	 * Also, add our configuration to it */
+	env_filter();
+	env_config();
+
 	newlevel = argv[optind++];
 	/* To make life easier, we only have the shutdown runlevel as
 	 * nothing really needs to know that we're rebooting.
 	 * But for those that do, you can test against RC_REBOOT. */
 	if (newlevel) {
+		if (rc_is_user() && strcmp(newlevel, RC_LEVEL_SYSINIT) == 0)
+			eerrorx("%s: the sysinit runlevel can't be used for user services.", applet);
+
 		if (strcmp(newlevel, "reboot") == 0) {
 			newlevel = UNCONST(RC_LEVEL_SHUTDOWN);
 			setenv("RC_REBOOT", "YES", 1);
 		}
 	}
 
+	atexit(cleanup);
+
 	/* Enable logging */
 	setenv("EINFO_LOG", "openrc", 1);
 
@@ -954,18 +959,15 @@ int main(int argc, char **argv)
 	/* Load our deptree */
 	if ((main_deptree = _rc_deptree_load(0, &regen)) == NULL)
 		eerrorx("failed to load deptree");
-	if (exists(RC_DEPTREE_SKEWED))
+
+	if (existsat(rc_dirfd(RC_DIR_SVCDIR), "clock-skewed"))
 		ewarn("WARNING: clock skew detected!");
 
 	/* Clean the failed services state dir */
 	clean_failed();
 
-	if (mkdir(RC_STOPPING, 0755) != 0) {
-		if (errno == EACCES)
-			eerrorx("%s: superuser access required", applet);
-		eerrorx("%s: failed to create stopping dir `%s': %s",
-		    applet, RC_STOPPING, strerror(errno));
-	}
+	if (mkdirat(rc_dirfd(RC_DIR_SVCDIR), "rc.stopping", 0755) != 0)
+		eerrorx("%s: failed to create stopping dir '%s/rc.stopping': %s", applet, rc_svcdir(), strerror(errno));
 
 	/* Create a list of all services which we could stop (assuming
 	* they won't be active in the new or current runlevel) including
@@ -1046,7 +1048,7 @@ int main(int argc, char **argv)
 	    going_down ? newlevel : runlevel);
 	hook_out = 0;
 
-	rmdir(RC_STOPPING);
+	unlinkat(rc_dirfd(RC_DIR_SVCDIR), "rc.stopping", AT_REMOVEDIR);
 
 	/* Store the new runlevel */
 	if (newlevel) {
@@ -1063,7 +1065,8 @@ int main(int argc, char **argv)
 		rc_logger_close();
 #endif
 
-	mkdir(RC_STARTING, 0755);
+	mkdirat(rc_dirfd(RC_DIR_SVCDIR), "rc.starting", 0755);
+
 	rc_plugin_run(RC_HOOK_RUNLEVEL_START_IN, runlevel);
 	hook_out = RC_HOOK_RUNLEVEL_START_OUT;
 
@@ -1133,7 +1136,7 @@ int main(int argc, char **argv)
 	 * default runlevel as they may depend on things that are now
 	 * available */
 	if (regen && strcmp(runlevel, bootlevel) == 0)
-		unlink(RC_DEPTREE_CACHE);
+		unlinkat(rc_dirfd(RC_DIR_SVCDIR), "deptree", 0);
 
 	return EXIT_SUCCESS;
 }
